#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <stdarg.h>
#include "config.h"

// TODO: use enums Generic, Specific, Virtual, Unset, etc.
//			then "gen" and "virt" can be renamed to "type"
// TODO: move * to end of line in input format?
// TODO: try making children a linked list

enum {
	Maxline = 4096,
	Maxname = 2048,
	Nlinks = 94,	/* non-whitespace ascii characters from '!' to '~' */
};

/*
* using char instead of int for val, gen, and virt cuts memory
* usage by about 15%.  could pack bits to save another 3-5%,
* though they are not strictly booleans (can hold -1).
*/
typedef struct Tdata Tdata;
struct Tdata {
	char *name;
	char val[Nattrs][Nsources];
	char gen[Nattrs][Nsources];	/* 1 if generic (prefix) assertion, 0 if specific */
	char virt[Nattrs][Nsources];	/* 1 if due to a generic assertion for a prefix of name */
	float exp[Nattrs];
	float nvexp[Nattrs];	/* exp from non-virtual assertions only */
};

typedef struct Trie Trie;
struct Trie {
	Tdata *dat;
	Trie *link[Nlinks];
};

/* generic assertions being propagated */
typedef struct Ginfo Ginfo;
struct Ginfo {
	char val[Nattrs][Nsources];
};

Trie *root;
int curtier;
float rep[Nsources];

Tdata *mktdata(char *);
void trieput(char *, int, int, int, int);
void printtrie(Trie *);
void setvirt();
void setrep();
void setexp();
void weprintf(const char *fmt, ...);

int
main()
{
	/* sources (resp. attrs) identified by ID between 0 and Nsources (resp. Nattrs) */
	int n, source, attr, value, generic, j;
	char line[Maxline], name[Maxname], *p;

	root = calloc(1, sizeof(Trie));

	while (fgets(line, Maxline, stdin) != NULL) {
		/* sscanf treats all whitespace the same; could look specifically for tabs */
		/* strtok probably faster anyway */
		p = line;
		generic = (*p == '*');
		p += generic;
		n = sscanf(p, "%d %d %d %s", &source, &attr, &value, &name);
		// XXX do proper input validation
		assert(source >= 0 && source < Nsources);
		assert(attr >= 0 && attr < Nattrs);
		assert(value == 0 || value == 1);
		trieput(name, attr, source, value, generic);
	}

	setvirt();
	for (curtier = 0; curtier < Ntiers; curtier++) {
		setrep();
		setexp();
	}
	printf("source\treputation\n");
	for (j = 0; j < Nsources; j++)
		printf("%d\t%f\n", j, rep[j]);
}

Tdata *
mktdata(char *name)
{
	Tdata *td;
	int i, j;
	td = calloc(1, sizeof(Tdata));
	td->name = strdup(name);
	for (i = 0; i < Nattrs; i++) {
		for (j = 0; j < Nsources; j++)
			td->val[i][j] = -1;
		td->exp[i] = td->nvexp[i] = -1;
	}
	return td;
}

Trie **
tlink(Trie *t, char c)
{
	assert (c >= '!' && c <= '~'); // XXX do proper input validation
	return t->link + (c - '!');
}

void
trieput(char *k, int attr, int source, int value, int generic)
{
	Trie *t, *new;
	char *p;
	t = root;
	p = k;

	for (;;) {
		// TODO: maintain skip count to save memory?  or convert to binary trie...
		if (t->dat != NULL && t->dat->name[p - k] != 0) {
			/* existing name not at end
			  * so t->dat was placed in the last if clause of this loop
			  * so there should be no further links */
			int i;
			for (i = 0; i < Nlinks; i++)
				assert(t->link[i] == NULL); // XXX

			//if (strcmp(k, t->dat->name) == 0)
			/* equivalent but faster: */
			if (strcmp(p, t->dat->name + (p - k)) == 0)
				break;
			else {
				/* out of my way! */
				new = calloc(1, sizeof(Trie));
				new->dat = t->dat;
//				t->link[t->dat->name[p-k]] = new;
				*tlink(t, t->dat->name[p-k]) = new;
				t->dat = NULL;
			}
 		}
		if (*p == 0) {
			if (t->dat == NULL)
				t->dat = mktdata(k);
			break;
		}
		if (*tlink(t, *p) == NULL) {
			new = calloc(1, sizeof(Trie));
			new->dat = mktdata(k);
//			t->link[*p] = new;
			*tlink(t, *p) = new;
			t = new;
			break;
		}
		t = *tlink(t, *p);
		++p;
	}
	assert(strcmp(t->dat->name, k) == 0); //XXX
	t->dat->val[attr][source] = value;
	t->dat->gen[attr][source] = generic;
}

void
_setvirt(Trie *t, Ginfo g)
{
	int i, j;

	if (t->dat != NULL) for (i = 0; i < Nattrs; i++) for (j = 0; j < Nsources; j++) {
		if (t->dat->val[i][j] < 0) {
			if (g.val[i][j] >= 0) {
				t->dat->val[i][j] = g.val[i][j];
				t->dat->virt[i][j] = 1;
			}
		} else if (t->dat->gen[i][j]) {
			if (g.val[i][j] == t->dat->val[i][j])
				weprintf("redundant assertion: source %d, attr %d, name %s\n",
						j, i, t->dat->name);
			else if (g.val[i][j] == !t->dat->val[i][j])
				weprintf("contradictory assertion: source %d, attr %d, name %s\n",
						j, i, t->dat->name);
			g.val[i][j] = t->dat->val[i][j];
		}
	}

	for (i = 0; i < Nlinks; i++)
		if (t->link[i] != NULL)
			_setvirt(t->link[i], g);	/* array is copied */
}

void setvirt(void)
{
	int i, j;
	Ginfo g;

	for (i = 0; i < Nattrs; i++)
		for (j = 0; j < Nsources; j++)
			g.val[i][j] = -1;
	_setvirt(root, g);
}

void
_setrep(Trie *t, float n[Nsources], float d[Nsources])
{
	int i, j, v, x;
	float c, e;
	Tdata *td;

	if ((td = t->dat) != NULL) {
		for (i = 0; i < Nattrs; i++) for (j = 0; j < Nsources; j++) {
			if (tier[j] == curtier && td->val[i][j] >= 0) {
				/*
				* we only compare virtual assertions against non-virtual
				* assertions from higher tiers.  this prevents lower
				* sources from affecting the weight given to an agreement
				* or disagreement between generic assertions by
				* adding assertions covered by both.  this way, reputations
				* depend only on assertions from higher tiers and not on
				* the total number of assertions we happen to have for a
				* certain name prefix.
				*/
				if (td->virt[i][j])
					e = td->nvexp[i];
				else
					e = td->exp[i];
				if (e < 0 || e == 0.5)
					continue;
				x = e < 0.5 ? 0 : 1;
				c = 2*fabsf(e - 0.5);
				/* weight by certainty unlike last paper */
				n[j] += c*abs(x - td->val[i][j]);	/* 1 if disagree */
				d[j] += c;
			}
		}
	}
	for (i = 0; i < Nlinks; i++)
		if (t->link[i] != NULL)
			_setrep(t->link[i], n, d);
}

void
setrep() {
	float n[Nsources], d[Nsources], x;
	int j;

	for (j = 0; j < Nsources; j++)
		n[j] = d[j] = 0;

	if (curtier != 0)
		_setrep(root, n, d);
	for (j = 0; j < Nsources; j++) {
		if (tier[j] == curtier) {
			if (d[j] == 0)
				rep[j] = 1; /* 0 for nonzero tiers in last paper */ //not sure which is better
			else
				rep[j] = (x = 1 - 2*(n[j] / d[j])) > 0 ? x : 0;
		}
	}
}

void
_setexp(Trie *t) {
	int i, j, v;
	Tdata *td;
	float n, d, nvn, nvd;

	if ((td = t->dat) != NULL) {
		for (i = 0; i < Nattrs; i++) {
			if (td->nvexp[i] >= 0)	/* implies t->exp[i] >= 0 */
				continue;	/* already set by higher tier */
			n = d = nvn = nvd = 0;
			for (j = 0; j < Nsources; j++) {
				if (tier[j] == curtier && td->val[i][j] >= 0) {
					n += rep[j] * td->val[i][j];
					d += rep[j];
					if (!td->virt[i][j]) {
						nvn += rep[j] * td->val[i][j];
						nvd += rep[j];
					}
				}
			}
			if (d != 0 && td->exp[i] < 0)
				td->exp[i] = n / d;
			if (nvd != 0)
				td->nvexp[i] = nvn / nvd;
		}
	}
	for (i = 0; i < Nlinks; i++)
		if (t->link[i] != NULL)
			_setexp(t->link[i]);
}

void
setexp()
{
	_setexp(root);
}

void
printtrie(Trie *t)
{
	int i, j, v;
	Tdata *td;
	if ((td = t->dat) != NULL) {
		printf("%s\n", td->name);
		for (i = 0; i < Nattrs; i++) {
			for (j = 0; j < Nsources; j++)
				if (td->val[i][j] >= 0) {
					printf("	attr %d:	%d says %d (gen: %d, virt: %d)\n", i, j, td->val[i][j], td->gen[i][j], td->virt[i][j]);
				}
			if (t->dat->exp[i] >= 0)
				printf("		exp: %f\n", td->exp[i]);
			if (t->dat->nvexp[i] >= 0)
				printf("		nvexp: %f\n", td->nvexp[i]);
		}
	}
	for (i = 0; i < Nlinks; i++)
		if (t->link[i] != NULL)
			printtrie(t->link[i]);
}

void
weprintf(const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
}
